#!/usr/bin/env python
#
# CDPreferences - add-on preferences QDialog for CellDraw - Mitja 2010
#
# ------------------------------------------------------------

from PyQt4 import QtCore, QtGui


from PyQt4.QtGui import *
from PyQt4.QtCore import *


# 2011 - Mitja: external class defining all global constants for CellDraw:
from cdConstants import CDConstants

# 2011 - Mitja: import classes generating bound signals
#        (from objects i.e. class instances, not from classes)
#        that change global preferences:
#
# to refer to the signalPIFFGenerationModeHasChanged signal
#   defined in CDControlPIFFGenerationMode we need the specific instance:
from cdTableOfTypes import CDTableOfTypes



# ----------------------------------------------------------------------
# 2010- Mitja: this is the main class for CellDraw preferences,
#   and it implements the dialog widget as well
# ----------------------------------------------------------------------
# note: this class emits a signal:
#
#         self.emit(QtCore.SIGNAL("cdPreferencesChangedSignal()"))
#
class CDPreferences(QtGui.QDialog):

    # ---------------------------------------------------------
    def __init__(self, pParent=None):
    # ---------------------------------------------------------
        # it is compulsory to call the parent's __init__ class right away:
        super(CDPreferences, self).__init__(pParent)
       
        self.theMainWindow = pParent

        #  __init__ (1) - set up the application's identity and credentials:
        # ---------------------------------------------------------
        #
        QtCore.QCoreApplication.setApplicationName("CellDraw");
        QtCore.QCoreApplication.setApplicationVersion("1.5.0");
        QtCore.QCoreApplication.setOrganizationName("Biocomplexity");
        QtCore.QCoreApplication.setOrganizationDomain("compucell.org");

        #  __init__ (2) - assign default values to all stored preferences,
        #     these values are used when there's nothing stored yet in the preferences file:
        # ---------------------------------------------------------
        #
        self.pifSceneWidth  = 320
        self.pifSceneHeight = 320
        self.pifSceneDepth  = 1
        self.pifSceneUnits  = "Pixel"
        self.pifScenePossibleUnitsList = ["Pixel", "mm", "micron"]
       
        # the class global keeping track of the selected PIFF generation mode:
        #    CDConstants.PIFFSaveWithFixedRaster,
        #    CDConstants.PIFFSaveWithOneRasterPerRegion,
        #    CDConstants.PIFFSaveWithPotts = range(3)
        self.piffGenerationMode = CDConstants.PIFFSaveWithOneRasterPerRegion
        self.piffFixedRasterWidth = 10

        #
        # if there is a preferences file already, then read from it:
        self.readPreferencesFromDisk()



        #  __init__ (4) - set up the dialog widget's GUI,
        # ---------------------------------------------------------
        #

        #  create a placeholder widget for preferences that are going to be added
        #    from the main CellDraw code:
        self.morePreferencesWidget = QtGui.QWidget()
        # print "___ - DEBUG ----- CDPreferences: __init__(): A"


        #  create a widget and layout for Cell Scene dimensions preferences:
        self.pifSceneWidthLineEdit = self.createLineEdit(str(self.pifSceneWidth), self.setPifSceneWidth)
        self.pifSceneHeightLineEdit = self.createLineEdit(str(self.pifSceneHeight), self.setPifSceneHeight)
        self.pifSceneDepthLineEdit = self.createLineEdit(str(self.pifSceneDepth), self.setPifSceneDepth)
        # print "___ - DEBUG ----- CDPreferences: __init__(): B"

        self.pifSceneUnitsComboBox = self.createComboBox(self.pifScenePossibleUnitsList, self.setPifSceneUnits)
        # print "___ - DEBUG ----- CDPreferences: __init__(): C"

        lFont = QtGui.QFont()
        lFont.setWeight(QtGui.QFont.Bold)
        pifSceneDimensionsTitleLabel = QtGui.QLabel("Cell Scene Dimensions")
        pifSceneDimensionsTitleLabel.setFont(lFont)
        pifSceneDimensionsTitleLabel.setMargin(5)
        pifSceneDimensionsTitleLabel.setFrameShape(QtGui.QFrame.Panel)
        pifSceneDimensionsTitleLabel.setPalette(QtGui.QPalette(QtGui.QColor(QtCore.Qt.lightGray)))
        pifSceneDimensionsTitleLabel.setAutoFillBackground(True)
        # print "___ - DEBUG ----- CDPreferences: __init__(): D"
       
        pifSceneWidthLabel = QtGui.QLabel("Cell Scene Width (x) :")
        pifSceneHeightLabel = QtGui.QLabel("Cell Scene Height (y) :")
        pifSceneDepthLabel = QtGui.QLabel("Cell Scene Depth (z) :")
        pifSceneUnitsLabel = QtGui.QLabel("Cell Scene Units:")

        placeHolderWidget = QtGui.QWidget()
        placeHolderWidget.setPalette(QtGui.QPalette(QtGui.QColor(QtCore.Qt.lightGray)))
        placeHolderWidget.setAutoFillBackground(True)

        sceneDimensionsLayout = QtGui.QGridLayout()

        sceneDimensionsLayout.addWidget(pifSceneDimensionsTitleLabel, 0, 0)

        sceneDimensionsLayout.addWidget(pifSceneWidthLabel, 1, 0)
        sceneDimensionsLayout.addWidget(self.pifSceneWidthLineEdit, 1, 1)

        sceneDimensionsLayout.addWidget(pifSceneHeightLabel, 2, 0)
        sceneDimensionsLayout.addWidget(self.pifSceneHeightLineEdit, 2, 1)

        sceneDimensionsLayout.addWidget(pifSceneDepthLabel, 3, 0)
        sceneDimensionsLayout.addWidget(self.pifSceneDepthLineEdit, 3, 1)

        sceneDimensionsLayout.addWidget(pifSceneUnitsLabel, 4, 0)
        sceneDimensionsLayout.addWidget(self.pifSceneUnitsComboBox, 4, 1)

        sceneDimensionsLayout.addWidget(placeHolderWidget, 5, 0)

        sceneDimensionsWidget = QtGui.QWidget()
        sceneDimensionsWidget.setLayout(sceneDimensionsLayout)
        sceneDimensionsWidget.setPalette(QtGui.QPalette(QtGui.QColor(QtCore.Qt.lightGray)))
        sceneDimensionsWidget.setAutoFillBackground(True)
       
        mainPlaceHolderWidget = QtGui.QWidget()


        #  create the main layout for the CDPreferences QDialog:
        self.mainLayout = QtGui.QGridLayout()
        self.mainLayout.addWidget(sceneDimensionsWidget,      0, 0, 1, 2)
        self.mainLayout.addWidget(mainPlaceHolderWidget,      1, 0, 1, 2)       
        self.mainLayout.addWidget(self.morePreferencesWidget, 0, 2, 2, 1)
        self.cancelButton = self.createButton("Cancel", self.readPreferencesFromDiskAndCloseDialog)
        self.savePreferencesButton = self.createButton("Save Preferences", self.writePreferencesToDiskAndCloseDialog, True)
        self.mainLayout.addWidget(self.cancelButton, 2, 0)
        self.mainLayout.addWidget(self.savePreferencesButton, 2, 2)
        self.setLayout(self.mainLayout)

        self.setWindowTitle("CellDraw - Preferences")
        #self.resize(700, 300)
        # self.setMinimumSize(400, 180)


        # this panel is a dialog "Window" (by PyQt and Qt definitions) :
        miDialogsWindowFlags = QtCore.Qt.WindowFlags()
        # this panel is a so-called "Tool" (by PyQt and Qt definitions)
        #    we'd use the Tool type of window, except for this oh-so typical Qt bug:
        #    http://bugreports.qt.nokia.com/browse/QTBUG-6418
        #    i.e. it defines a system-wide panel which shows on top of *all* applications,
        #    even when this application is in the background.
        # miDialogsWindowFlags = QtCore.Qt.Tool
        #    so we use a plain QtCore.Qt.Dialog type instead:
        miDialogsWindowFlags = QtCore.Qt.Dialog

        #    add a peculiar WindowFlags combination to have no close/minimize/maxize buttons:
        miDialogsWindowFlags |= QtCore.Qt.WindowTitleHint
        miDialogsWindowFlags |= QtCore.Qt.CustomizeWindowHint
#        miDialogsWindowFlags |= QtCore.Qt.WindowMinimizeButtonHint
#        miDialogsWindowFlags |= QtCore.Qt.WindowStaysOnTopHint
        self.setWindowFlags(miDialogsWindowFlags)


        # 1. The widget is not modal and does not block input to other widgets.
        # 2. If widget is inactive, the click won't be seen by the widget.
        #    (it does NOT work as Qt docs says it would on Mac OS X: click-throughs don't get disabled)
        # 3. The widget can choose between alternative sizes for widgets to avoid clipping.
        # 4. The native Carbon size grip should be opaque instead of transparent.
        self.setAttribute(QtCore.Qt.NonModal  | \
                          QtCore.Qt.WA_MacNoClickThrough | \
                          QtCore.Qt.WA_MacVariableSize | \
                          QtCore.Qt.WA_MacOpaqueSizeGrip )

        # do not delete the window widget when the window is closed:
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, False)

        print "___ - DEBUG ----- CDPreferences: __init__(): done"


    # ---------------------------------------------------------
    # add a QWidget within the preferences dialog:
    # ---------------------------------------------------------
    def addMorePreferencesWidget(self, pTheWidget=None):
        if pTheWidget is not None:
            # remove the previous self.morePreferencesWidget in the layout:
            self.mainLayout.removeWidget(self.morePreferencesWidget)
            # the above "removeWidget" statement does not seem to be enough. One must hide the widget too:
            self.morePreferencesWidget.hide()

            # add the new self.morePreferencesWidget widget to the layout:
            self.morePreferencesWidget = None # <----- do we need this ???
            self.morePreferencesWidget = pTheWidget
            self.mainLayout.addWidget(self.morePreferencesWidget, 0, 2, 2, 1)

            self.mainLayout.update()
       
           
    # ---------------------------------------------------------
    # decide whether to show the "more preferences" QWidget in the dialog:
    # ---------------------------------------------------------
    def showMorePrefs(self, pShowMorePrefs=True):
        if pShowMorePrefs is False:
            self.mainLayout.removeWidget(self.morePreferencesWidget)
            self.morePreferencesWidget.hide()
            self.mainLayout.update()
        else:
            self.morePreferencesWidget.show()
            self.mainLayout.addWidget(self.morePreferencesWidget, 0, 2, 2, 1)
            self.mainLayout.update()


    # ---------------------------------------------------------
    # the reject() function is called when the user presses the <ESC> keyboard key.
    #   We override the default and implicitly include the equivalent action of
    #   <esc> being the same as clicking the "Cancel" button, as in well-respected GUI paradigms:
    # ---------------------------------------------------------
    def reject(self):
        self.cancelButton.click()
        super(CDPreferences, self).reject()
        # print "reject() DONE"


    # ---------------------------------------------------------
    def readPreferencesFromDiskAndCloseDialog(self):
        self.readPreferencesFromDisk()
        self.close()
        # self.hide()
        # print "readPreferencesFromDiskAndCloseDialog() DONE"
   

    # ---------------------------------------------------------
    def readPreferencesFromDisk(self):
        preferences = QSettings(QtCore.QCoreApplication.organizationName(), QtCore.QCoreApplication.applicationName())
        preferences.beginGroup("pifScene")
        # each value saved in the preferences has to be extracted from a QVariant object,
        #   and if there is no such preferences key/value yet,
        #   QVariant builds it from the default values we place there (our globals):
        # print "self.pifSceneWidth, self.pifSceneHeight, self.pifSceneDepth, self.pifSceneUnits =", self.pifSceneWidth, self.pifSceneHeight, self.pifSceneDepth, self.pifSceneUnits
        # as we read int values from preferences, .value() returns them as tuples:
        self.pifSceneWidth, ok  = preferences.value("pifSceneWidth", \
                                QVariant(self.pifSceneWidth) ).toInt()
        self.pifSceneHeight, ok = preferences.value("pifSceneHeight", \
                                QVariant(self.pifSceneHeight) ).toInt()
        self.pifSceneDepth, ok = preferences.value("pifSceneDepth", \
                                QVariant(self.pifSceneDepth) ).toInt()
        self.pifSceneUnits  = preferences.value("pifSceneUnits", \
                                QVariant(self.pifSceneUnits) ).toString()
        # print "self.pifSceneWidth, self.pifSceneHeight, self.pifSceneDepth, self.pifSceneUnits =", self.pifSceneWidth, self.pifSceneHeight, self.pifSceneDepth, self.pifSceneUnits
        preferences.endGroup()
        # print "readPreferencesFromDisk() DONE"


    # ---------------------------------------------------------
    def writePreferencesToDiskAndCloseDialog(self):
        self.writePreferencesToDisk()
        self.hide()
   

    # ---------------------------------------------------------
    def writePreferencesToDisk(self):
        preferences = QSettings(QtCore.QCoreApplication.organizationName(), QtCore.QCoreApplication.applicationName())
        preferences.beginGroup("pifScene")
        # each value saved in the preferences has to be wrapped inside a QVariant object:
        preferences.setValue("pifSceneWidth", QVariant(self.pifSceneWidth))
        preferences.setValue("pifSceneHeight", QVariant(self.pifSceneHeight))
        preferences.setValue("pifSceneDepth", QVariant(self.pifSceneDepth))
        preferences.setValue("pifSceneUnits", QVariant(self.pifSceneUnits))
        preferences.endGroup();

        # propagate a signal upstream, for example to parent objects:
        self.emit(QtCore.SIGNAL("cdPreferencesChangedSignal()"))

        # print "writePreferencesToDisk() DONE"


    # ---------------------------------------------------------
    def setPifSceneWidth(self):
        self.pifSceneWidth = int(self.pifSceneWidthLineEdit.text())

    # ---------------------------------------------------------
    def setPifSceneHeight(self):
        self.pifSceneHeight = int(self.pifSceneHeightLineEdit.text())

    # ---------------------------------------------------------
    def setPifSceneDepth(self):
        self.pifSceneDepth = int(self.pifSceneDepthLineEdit.text())

    # ---------------------------------------------------------
    def setPifSceneUnits(self):
        self.pifSceneUnits = str(self.pifSceneUnitsComboBox.currentText())
        # print "pifSceneUnits =", self.pifSceneUnits

    # ---------------------------------------------------------
    def createLineEdit(self, text, callback, isdefault=False):
        lineEdit = QtGui.QLineEdit(text)
        lineEdit.setInputMask("00000")
        lineEdit.editingFinished.connect(callback)
        return lineEdit


    # ---------------------------------------------------------
    def createButton(self, text, callback, isdefault=False):
        button = QtGui.QPushButton(text)
        button.setDefault(isdefault)
        button.clicked.connect(callback)
        return button

    # ---------------------------------------------------------
    def createComboBox(self, textList, callback, iseditable=False):
        comboBox = QtGui.QComboBox()
        for lText in textList:
            comboBox.addItem(lText)
        # comboBox.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Preferred)
        comboBox.setEditable(iseditable)
        # make certain that the signal-callback connection is made as *last* after having populated the widget,
        #    otherwise the callback would be called on a half-finished setup (the actual object is not available here yet!)
        comboBox.currentIndexChanged.connect(callback)
        return comboBox



    # --------------------------------------------------------------------
    # 2011 Mitja - connectSignalsToHandlers() has to be called after
    #              all other classes have been initialized, otherwise the signals
    #              may not have been defined yet!
    # --------------------------------------------------------------------
    def connectSignalsToHandlers(self):   
        # explicitly connect the "signalLayersSelectionModeHasChanged()"
        #   signal from the theControlsForLayerSelection object,
        #   to our "slot" method responding to radio button changes:
        self.theMainWindow.theTableOfTypes.theControlsForPIFFGenerationMode.signalPIFFGenerationModeHasChanged.connect( \
            self.handlePIFFGenerationModeHasChanged )


    # ------------------------------------------------------------------
    # 2011 Mitja - slot method handling "signalLayersSelectionModeHasChanged"
    #    events (AKA signals) arriving from  theControlsForLayerSelection:
    # ------------------------------------------------------------------
    def handlePIFFGenerationModeHasChanged(self, pNewMode, pNewFixedRasterSize):
        if (pNewMode != self.piffGenerationMode) or (pNewFixedRasterSize != self.piffFixedRasterWidth):
            print "CDPreferences() - handlePIFFGenerationModeHasChanged(pNewMode=",pNewMode,", pNewFixedRasterSize=",pNewFixedRasterSize,")"
            # change global variables AKA preferences:
            self.piffGenerationMode = pNewMode
            self.piffFixedRasterWidth = pNewFixedRasterSize
            # automatically save preferences to disk:
            self.writePreferencesToDisk()


if __name__ == '__main__':

    import sys

    app = QtGui.QApplication(sys.argv)
    window = CDPreferences()
    window.show()
    window.raise_()
    sys.exit(app.exec_())


# Local Variables:
# coding: US-ASCII
# End:
